/*
    Copyright (c) 2002-2009 Tampere University of Technology.

    This file is part of TTA-Based Codesign Environment (TCE).

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
 */
/**
 * @file DataDependenceGraph.hh
 *
 * Declaration of data dependence graph class
 *
 * @author Heikki Kultala 2006-2009 (heikki.kultala-no.spam-tut.fi)
 * @note rating: red
 */

#ifndef TTA_DATA_DEPENDENCE_GRAPH_HH
#define TTA_DATA_DEPENDENCE_GRAPH_HH

#include "BoostGraph.hh"
#include "DataDependenceEdge.hh"

#include <map>
#include <set>
#include <list>
#include <utility>
#include <vector>
#include <utility>

#include "MoveNode.hh"

typedef std::vector<class ProgramOperation*> POList;
typedef POList::iterator POLIter;

class BasicBlockNode;
class BasicBlock;
class DataGraphBuilder;
class ControlFlowGraph;
class MoveNodeUse;

namespace TTAMachine {
    class BaseRegisterFile;
    class Machine;
}

namespace TTAProgram {
    class CodeSnippet;
}

class DataDependenceGraph : 
    public BoostGraph<MoveNode, DataDependenceEdge> {
public:

    enum AntidependenceLevel {
        NO_ANTIDEPS,
        INTRA_BB_ANTIDEPS,
        SINGLE_BB_LOOP_ANTIDEPS,
        ALL_ANTIDEPS
    };

    enum DumpFileFormat {
        DUMP_DOT,
        DUMP_XML
    };

    enum EdgeWeightHeuristics {        
        EWH_HEURISTIC,   ///< Weights memory dependencies more, etc.
        EWH_REAL,        ///< Height returns the minimum cycle count for the
                         ///< schedule given enough resources.
        EWH_DEFAULT = EWH_HEURISTIC
    };

    DataDependenceGraph(
        std::set<TCEString> allParamRegs,
        const TCEString& name="", 
        AntidependenceLevel antidependenceLevel = ALL_ANTIDEPS,
        BasicBlockNode* ownedBBN = NULL, 
        bool containsProcedure=false, bool noLoopEdges=false);

    virtual ~DataDependenceGraph();    

    /* Gets the BB in which this move belongs. This should be moved to
       a better place: DDG API should not deal with basic blocks. Correct
       place BasicBlockNode? */
    const BasicBlockNode& getBasicBlockNode(const MoveNode& mn) const;
    BasicBlockNode& getBasicBlockNode(MoveNode& mn);

    int programOperationCount();
    ProgramOperation& programOperation(int index);

    DataDependenceEdge* onlyRegisterEdgeIn(MoveNode& mn) const;
    DataDependenceEdge* onlyRegisterEdgeOut(MoveNode& mn) const;

    // should not be called by the user, only by a DataDependenceGraphBuilder
    void addNode(MoveNode& moveNode) 
        throw (ObjectAlreadyExists);
    void addNode(MoveNode& moveNode, MoveNode& relatedNode);
    void addNode(MoveNode& moveNode, BasicBlockNode& bblock);
    void addProgramOperation(ProgramOperation* po);

    using BoostGraph<MoveNode, DataDependenceEdge>::addNode;

    int earliestCycle(
        const MoveNode& moveNode, unsigned int ii = UINT_MAX, 
        bool ignoreRegWaRs = false, 
        bool ignoreRegWaWs = false, bool ignoreGuards = false,
        bool ignoreFUDeps = false) const;

    int latestCycle(
        const MoveNode& moveNode, unsigned int ii = UINT_MAX,
        bool ignoreRegAntideps = false) const;
    
    int smallestCycle() const;
    int largestCycle() const;
    int scheduledNodeCount() const;

    NodeSet unscheduledMoves() const;
    NodeSet scheduledMoves() const;
    NodeSet movesAtCycle(int cycle) const;

    MoveNode* lastGuardDefMove(MoveNode& moveNode);
    NodeSet guardDefMoves(const MoveNode& moveNode);

    MoveNode*
    lastScheduledRegisterRead(
        const TTAMachine::BaseRegisterFile& rf, int registerIndex) const;

    NodeSet
    lastScheduledRegisterReads(
        const TTAMachine::BaseRegisterFile& rf, int registerIndex) const;

    NodeSet
    lastScheduledRegisterGuardReads(
        const TTAMachine::BaseRegisterFile& rf, int registerIndex) const;
    
    NodeSet
    lastScheduledRegisterWrites(
        const TTAMachine::BaseRegisterFile& rf, int registerIndex) const;

    MoveNode*
    lastScheduledRegisterKill(
        const TTAMachine::BaseRegisterFile& rf, int registerIndex) const;

    int lastRegisterCycle(
        const TTAMachine::BaseRegisterFile& rf, int registerIndex) const;

    void sanityCheck() const
        throw (Exception);

    /// Dot printing related methods
    virtual TCEString dotString() const;
    virtual void setCycleGrouping(bool flag);

    // XML dumping
    void writeToXMLFile(std::string fileName) const;

    bool mergeAndKeep(MoveNode& resultNode, MoveNode& userNode);
    void unMerge(MoveNode& resultNode, MoveNode& mergedNode);

    bool resultUsed(MoveNode& resultNode);
    
    void removeNode(MoveNode& node) throw (InstanceNotFound);
    void deleteNode(MoveNode& node);

    void setEdgeWeightHeuristics(EdgeWeightHeuristics ewh) {
        if (edgeWeightHeuristics_ != ewh) {
            // force path recalculation
            height_ = -1; 
            sourceDistances_.clear();
            sinkDistances_.clear();
        }
        edgeWeightHeuristics_ = ewh;
    }

    int edgeWeight(DataDependenceEdge& e, const MoveNode& hNode) const;
    bool predecessorsReady(MoveNode& node) const;
    void setMachine(const TTAMachine::Machine& machine);
    const TTAMachine::Machine& machine() const { return *machine_; } 

    bool connectOrDeleteEdge(        
        const MoveNode& tailNode, const MoveNode& headNode, 
        DataDependenceEdge* edge);

    DataDependenceGraph* createSubgraph(
        NodeSet& nodes, bool includeLoops = false);

    DataDependenceGraph* createSubgraph(
        TTAProgram::CodeSnippet& cs, bool includeLoops = false);

    DataDependenceGraph* createSubgraph(
        std::list<TTAProgram::CodeSnippet*>& codeSnippets,
        bool includeLoops = false);

    DataDependenceGraph* trueDependenceGraph(
        bool removeMemAntideps=true, bool ignoreMemDeps=false);
    DataDependenceGraph* criticalPathGraph();

    MoveNode& nodeOfMove(TTAProgram::Move& move) throw (InstanceNotFound);

    void dropBackEdges();

    void fixInterBBAntiEdges(BasicBlockNode& bbn1, 
                             BasicBlockNode& bbn2,
                             bool loopEdges)
        throw (Exception);

    // Duplicates all in- and outgoing edges in dst to src
    void copyDependencies(MoveNode& src, MoveNode& dst) 
        throw (InstanceNotFound); 

    void copyIncomingGuardEdges(const MoveNode& src, MoveNode& dst);
    void copyOutgoingGuardWarEdges(const MoveNode& src, MoveNode& dst);
    
    NodeSet guardRawPredecessors(const MoveNode& node) const;

    DataDependenceEdge* onlyIncomingGuard(const MoveNode& mn) const;
    MoveNode* onlyRegisterRawSource(const MoveNode& mn) const;

    NodeSet regWarSuccessors(const MoveNode& node) const;
    NodeSet regRawSuccessors(const MoveNode& node) const;
    NodeSet regWawSuccessors(const MoveNode& node) const;

    void createRegisterAntiDependenciesBetweenNodes(NodeSet& nodes);

    bool exclusingGuards(const MoveNode& mn1, const MoveNode& mn2) const;
    bool sameGuards(const MoveNode& mn1, const MoveNode& mn2) const;

    int rWarEdgesOut(MoveNode& mn);
    int regRawSuccessorCount(const MoveNode& mn, bool onlyScheduled);

    bool guardsAllowBypass(const MoveNode& defNode, const MoveNode& useNode);

    const MoveNode* onlyRegisterRawAncestor(
        const MoveNode& node, const std::string& sp) const;

    void copyDepsOver(MoveNode& node, bool anti, bool raw);

    TTAProgram::Move* findLoopLimit(MoveNode& jumpMove);

    void moveFUDependenciesToTrigger(MoveNode& trigger);

    std::pair<NodeSet,NodeSet> findLiveRange(
        MoveNode& lrNode, bool writingNode=true) const;

    MoveNode* findLimitingAntidependenceSource(MoveNode& mn);
    MoveNode* findLimitingAntidependenceDestination(MoveNode& mn);

    void sourceRenamed(MoveNode& mn);
    void destRenamed(MoveNode& mn);

    void copyExternalInEdges(MoveNode& nodeCopy, const MoveNode& source);
    void copyExternalOutEdges(MoveNode& nodeCopy, const MoveNode& source);

    void updateRegWrite(
        const MoveNodeUse& mnd, const TCEString& reg, BasicBlock& bb);

    void updateRegUse(
        const MoveNodeUse& mnd, const TCEString& reg, BasicBlock& bb);

    void removeIncomingGuardEdges(MoveNode& node);
    void removeOutgoingGuardWarEdges(MoveNode& node);

    inline bool hasAllRegisterAntidependencies() {
        return registerAntidependenceLevel_ >= ALL_ANTIDEPS;
    }

    inline bool hasSingleBBLoopRegisterAntidependencies() {
        return registerAntidependenceLevel_ >= SINGLE_BB_LOOP_ANTIDEPS;
    }

    inline bool hasIntraBBRegisterAntidependencies() {
        return registerAntidependenceLevel_ >= INTRA_BB_ANTIDEPS;
    }

private:
    bool rWawRawEdgesOutUncond(MoveNode& mn);
    int rAntiEdgesIn(MoveNode& mn);

    void setNodeBB(
        MoveNode& mn, BasicBlockNode& bblock, DataDependenceGraph* updater);

    bool isRootGraphProcedureDDG();
    
    bool hasEqualEdge(
        const MoveNode& tailNode, const MoveNode& headNode, 
        const DataDependenceEdge& edge) const;
    
    int getOperationLatency(const TCEString& name) const;

    std::set<TCEString> allParamRegs_;

    // cache to make things faster
    // may not be used with iterator.
    std::map<TTAProgram::Move*, MoveNode*> nodesOfMoves_;

    // own all the programoperations
    POList programOperations_;
    std::map<const MoveNode*, BasicBlockNode*> moveNodeBlocks_;

    /// Dot printing related variables.
    /// Group the printed MoveNodes according to their cycles.
    bool cycleGrouping_;

    // Machine related variables. 
    const TTAMachine::Machine* machine_;
    int delaySlots_;
    std::map<TCEString, int> operationLatencies_;

    BasicBlockNode* ownedBBN_;
    bool procedureDDG_;
    AntidependenceLevel registerAntidependenceLevel_;

    /// The heuristics to use to weight edges in the longest path computation.
    EdgeWeightHeuristics edgeWeightHeuristics_;
};

#endif
